#!/usr/bin/env python
# coding: utf-8

# In[1]:


import pandas as pd
import numpy as np
import math
import sys
from pandas.api.types import is_numeric_dtype
from pandas.api.types import is_string_dtype
import statsmodels.distributions.empirical_distribution as edf
from scipy.interpolate import interp1d
import inspect
import time
import warnings
warnings.filterwarnings("ignore")


# In[2]:


def var_cat(df):
    df_temp=df[[cols for cols in df.columns if cols not in y]]
    a=list(df_temp.apply(lambda x: 'binary' if len(x.dropna().unique())==2 else 'other'))
    binary_var=[x for x,t in zip(df_temp.columns,a) if t not in ['other']]
    binary_list=pd.DataFrame({'var_name':binary_var,'data_type':['binary']*len(binary_var)})
    del a
    
    b=list(df_temp.apply(lambda x: 'numeric' if len(x.dropna().unique())>0.5*len(x) else 'other'))
    numeric_var=[x for x,t in zip(df_temp.columns,b) if t not in ['other']]+[y]
    numeric_list=pd.DataFrame({'var_name':numeric_var,'data_type':['numeric']*len(numeric_var)})
    del b
    
    categorical_var=[cols for cols in df.columns if cols not in (binary_var+numeric_var)]
    categorical_list=pd.DataFrame({'var_name':categorical_var,'data_type':['categorical']*len(categorical_var)})  
    
    var_list=pd.concat([binary_list,numeric_list,categorical_list])
    
    return var_list


# In[4]:


def TnI_binary(bin_test,missing_threshold=0.01):
    '''
    Checks for columns with binary values, exlcuding nan and converts the column into dummy variables; output returned includes 
    bin_test: dataframe whose binary variables are to be converted
    missing_threshold: Conversion happens only if the number of missing values are greater than threshold percentage of total number of rows    
    Why...above a certain threshold?
    '''
    
    import pandas as pd
    bin_cols=bin_test.columns
    used_cols=[]
    
    #build lookup table with only one row and build a column indicating missing entries in x variable and also rename that column to x_missing_flag 
    lookup_table=pd.DataFrame()
    lookup_table_rep=pd.DataFrame()
    #lookup_table_bin=pd.DataFrame()
    
    for cols in list(bin_cols):
        if len(bin_test[cols].dropna().unique())==2:
            if sum(bin_test[cols].isna())>len(bin_test[cols])*missing_threshold:
                used_cols.append(cols)
                bin_test[cols]=bin_test[cols].fillna('missing')
                dummy_temp=pd.get_dummies(bin_test[cols],prefix=cols,dummy_na=False)
                #This code is added to get same features as in the training tranformation for binary as well
                count=0
                for x in dummy_temp.columns:
                    count=count+1
                    lookup_table=lookup_table.append({'var_name':x,'original_name':'_'.join(x.split('_')[0:len(x.split('_'))-1]),'type':'binary binary'},ignore_index=True)
                    lookup_table_rep=lookup_table_rep.append({'original_name':'_'.join(x.split('_')[0:len(x.split('_'))-1]),'old_val':x.split('_')[len(x.split('_'))-1],'new_val':count},ignore_index=True)                    
                bin_test=pd.concat([bin_test,dummy_temp],axis=1)
                del dummy_temp
                
            #This part I have added as I felt it was missing
            elif sum(bin_test[cols].isna())==0:
                used_cols.append(cols)
                dummy_temp=pd.get_dummies(bin_test[cols],prefix=cols,dummy_na=False)
                #This code is added to get same features as in the training tranformation for binary as well
                count=0
                for x in dummy_temp.columns:
                    count=count+1
                    lookup_table=lookup_table.append({'var_name':x,'original_name':'_'.join(x.split('_')[0:len(x.split('_'))-1]),'type':'binary binary'},ignore_index=True)
                    lookup_table_rep=lookup_table_rep.append({'original_name':'_'.join(x.split('_')[0:len(x.split('_'))-1]),'old_val':x.split('_')[len(x.split('_'))-1],'new_val':count},ignore_index=True)                    

                bin_test=pd.concat([bin_test,dummy_temp],axis=1)
                del dummy_temp
                
            else:
                sys.exit('There are small number of missing values in '+cols+'. Consider imputating it')
            
    bin_test=bin_test[[cols for cols in bin_test.columns if cols not in used_cols]]
    #testlist={'data':bin_test,'lookup':{'lookup_table':lookup_table,'lookup_table_bin':lookup_table_bin,'lookup_table_rep':lookup_table_rep}}
    testlist={'data':bin_test,'lookup':{'lookup_table':lookup_table,'lookup_table_rep':lookup_table_rep}}
    
    return testlist
    #return bin_test


# In[7]:


def TnI_cat(testdata, x, y, missing_flag = True, missing_thres = 0.5, min_bin_size = 0.1,             binary_var = True,bv_min_size = 0.1, bv_min_incindex = 0.1, bar = False):
       
    #Check if the dataset contains the column name
    if x not in testdata.columns: sys.exit('ERROR!!! x variable not found in dataset, please check variable name')
    if y not in testdata.columns: sys.exit('ERROR!!! y variable not found in dataset, please check variable name')
    

    #Create duplicate columns 
    testdata['x_royalias']=testdata[x]
    testdata['y_royalias']=testdata[y]
    
    #Exit with error for drop if missing values are more than a threshold percentage
    if sum(testdata['x_royalias'].isna())/len(testdata)>missing_thres: sys.exit('ERROR!!! x variable has more than '+str(round(missing_thres*100,1))+'% missing, please consider drop')
    
    #Exit with error if the column data type is not string
    if is_numeric_dtype(testdata['x_royalias']): sys.exit('ERROR!!! '+x+' variable is not string type. TnI_categorical requires x variable to be string, please check variable type')
    
    #Exit with error for drop if there is any missing value in y
    if sum(testdata['y_royalias'].isna())>0: sys.exit('ERROR!!! y (' +y+') variable has missing value. TnI_categorical needs y variable to be fully populated. Please consider eliminating missing y values')
    
    #Exit with error if the datatype for y is a string
    if is_string_dtype(testdata['y_royalias']): sys.exit('ERROR!!! y ('+y+') variable is not numeric or integer. TnI_categorical requires y variable to be numeric or integer, please check variable type')
    
    #Check if the variable has potential to be ids column
    if testdata['x_royalias'].nunique()>1000: sys.exit('ERROR!!! x variable is a character variable with more than 1000 levels, highly likely to be ID vars')

    #build lookup table with only one row and build a column indicating missing entries in x variable and also rename that column to x_missing_flag 
    lookup_table=pd.DataFrame({'var_name':x,'original_name':x,'type':'original'},index=[0])

    if missing_flag==True:
        if testdata['x_royalias'].isna().mean()>0: # do we need mean here or count would do?
            testdata['x_missing_flag']=[1 if ct==True else 0 for ct in list(pd.isnull(testdata['x_royalias']))]
            lookup_table=pd.concat([lookup_table,pd.DataFrame({'var_name':x+'_missing_flag','original_name':x,'type':'missing_flag'},index=[1])])
            testdata=testdata.rename(columns={'x_missing_flag':x+'_missing_flag'})
        else:
            if bar in [False]:
                print('there is no missing in independent variable, no inputing was done, no missing flag was created')


    #based on y_index to re-assign value
    testdata['x_royalias'].fillna('TnImissing',inplace=True) # to avoid issue when joining by this var
    testlookup=testdata[['y_royalias','x_royalias']].groupby(['x_royalias']).mean().reset_index()
    testlookup=testlookup.rename(columns={'y_royalias':'y_royindex1'})
    testlookup=testlookup.sort_values(by='y_royindex1',ascending=True)
    testlookup['x_royassign']=list(range(1,(len(testlookup)+1)))
    testlookup.drop(['y_royindex1'],axis=1,inplace=True)
    testlookup2=testlookup.copy()
    testlookup2['x_royalias'].replace('TnImissing',np.nan,inplace=True)
    lookup_table_rep=pd.DataFrame({'original_name':np.repeat(x,len(testlookup2)),'old_val':testlookup2['x_royalias'],'new_val':testlookup2['x_royassign']})


    #append back to testdata
    testdata=testdata.merge(testlookup,how='left',on='x_royalias')

    ## rename alias for lookup
    testdata['x_royalias'].replace('TnImissing',np.nan,inplace=True)
    testlookup=testlookup.rename(columns={'x_royalias':x}) ######Two columns with same name created...should I delete one?
    new_column_temp_name=x+'_TnI_assign'
    testlookup=testlookup.rename(columns={'x_royassign':new_column_temp_name})
    del new_column_temp_name
    
    
    ## bin the re-assigned value
    #Ignore divide by 0 warning in NumPy - globally
    np.seterr(divide='ignore')

    if bar == False:
        print('min_bin_size = '+str(min_bin_size))

    #Define later after building tab below
    if len(testlookup)==1:
        if bar ==False:
            print('With bounds, only one unique value')

    else:
        counter1=0
        counter2=0
        tab=testdata['x_royassign'].value_counts().sort_index()
        cut_point=np.full((len(tab)-1), np.nan)

        for ii in range(0,(len(tab)-1)):
            counter1=tab.iloc[ii]
            counter2=counter1+counter2
            if counter2/sum(testdata['x_royassign'].isna())>=min_bin_size:
                cut_point[ii]=int(tab.index[ii])
                counter2=0
        del counter1,counter2,ii
        cut_point=np.array([x for x in list(cut_point) if x is not np.nan])
        if sum(testdata['x_royassign']>cut_point[-1])/sum(testdata['x_royassign'].notnull())<min_bin_size:
            cut_point=cut_point[:-1]
    

    cut_point=[-np.inf]+list(cut_point)+[np.inf]

    testdata['x_roybins_bi']=pd.cut(testdata['x_royassign'],cut_point, right=True)
    testdata['x_roybins_bi']=testdata['x_roybins_bi'].astype(str)

    ## recreate y index for bins
    temp_df=testdata.groupby(['x_roybins_bi'])[['y_royalias']].mean().reset_index()
    temp_df=temp_df.rename(columns={'x_roybins_bi':'x_roybins_bi','y_royalias':'y_royindex'})
    testdata=testdata.merge(temp_df,how='left',left_on='x_roybins_bi',right_on='x_roybins_bi')


    ## take binary vars
    if binary_var==True:
        temp_x_roybins_bi=testdata['x_roybins_bi']
        testdata=pd.get_dummies(testdata,columns=['x_roybins_bi'])
        testdata['x_roybins_bi']=temp_x_roybins_bi
        del temp_x_roybins_bi
    
        drop2=[] 
    
        for ii in range((len(testdata.columns)-len(testdata['x_roybins_bi'].unique())-1),(len(testdata.columns)-1)):
            if testdata.iloc[:,ii].mean()<bv_min_size:
                drop2.append(ii)
            else:
                if abs(testdata['y_royindex'][testdata.iloc[:,ii]==1].mean()/testdata['y_royindex'].mean()-1)<bv_min_incindex:
                    drop2.append(ii)
    
        if len(drop2)==0:
            if bar == False:
                print('all binary variable taken are significant')
        else:
            if bar == False:
                print(str(len(testdata['x_roybins_bi'].unique())-len(drop2))+ ' significant binary variable was taken')
                testdata=testdata.drop(list(testdata.columns[drop2]),axis='columns')
        del drop2, ii
    
    ## rename alias
    lookup_table=lookup_table.append({'var_name':(x+'_TnI_assign'), 'original_name':x,'type':'assign'}, ignore_index=True)
    testdata=testdata.rename(columns={'x_royassign':(x+'_TnI_assign')})
    lookup_table=lookup_table.append({'var_name':(x+'_assign_bins'), 'original_name':x,'type':'assign_bins'}, ignore_index=True)
    testdata=testdata.rename(columns={'x_roybins_bi':(x+'_assign_bins')})
    idx=[testdata.columns.get_loc(cols) for cols in testdata.columns if cols.startswith('x_roybins')]
    lookup_table_bin=pd.DataFrame()

    temp_upper_bound=[]
    temp_lower_bound=[]
    temp_var_name=[]

    for ii in idx:
        temp_lower_bound.append(float(testdata.columns[ii].split(',')[0].split('(')[1]))
        temp_upper_bound.append(float(testdata.columns[ii].split(',')[1].replace(']','').replace(' ','')))
        temp_var_name.append(testdata.columns[ii].replace('x_roybins',x))
        lookup_table=lookup_table.append({'var_name':testdata.columns[ii].replace('x_roybins_bi_',(x+'_bi_')), 'original_name':x,'type':'categorical binary'}, ignore_index=True)
    
    lookup_table_bin['var_name']=temp_var_name
    lookup_table_bin['lower_bound']=temp_lower_bound
    lookup_table_bin['upper_bound']=temp_upper_bound
    testdata.columns=[cols.replace('x_roybins_bi_',(x+'_bi_')) for cols in testdata.columns]
    lookup_table=lookup_table.append({'var_name':(x+'_dependent_var_index'), 'original_name':x,'type':'dependent_var_index'}, ignore_index=True)
    testdata=testdata.rename(columns={'y_royindex':(x+'_dependent_var_index')})
    testdata=testdata.drop(['x_royalias','y_royalias'],axis='columns')

    testlist={'data':testdata,'lookup':{'lookup_table':lookup_table, 'lookup_table_bin':lookup_table_bin, 'lookup_table_rep':lookup_table_rep}}
    
    return testlist


# In[10]:


def quantile_function_1(a,bounds):
    '''Quantile based on Inverse of Empirical Distribution Function'''
    
    sample_edf = edf.ECDF(a)
    slope_changes = sorted(set(a))
    sample_edf_values_at_slope_changes = [sample_edf(item) for item in slope_changes]
    inverted_edf = interp1d(sample_edf_values_at_slope_changes, slope_changes)
    
    bnds_temp=[]
    bnds_temp.append(np.asscalar(inverted_edf(.5-bounds/2)))
    bnds_temp.append(np.asscalar(inverted_edf(.5+bounds/2)))
    
    return bnds_temp


# In[11]:


def quantile_function_2(a,bounds):
    '''Quantile based on Neqarest Neighbour'''
    
    bnds_temp=[]
    bnds_temp.append(np.quantile(a, q=(.5-bounds/2),interpolation='nearest'))
    bnds_temp.append(np.quantile(a, q=(.5+bounds/2),interpolation='nearest'))
    
    return bnds_temp


# In[12]:


def TnI_cont(testdata, x, y, missing_flag = True, missing_thres = 0.5, min_bin_size = 0.1, method = 'index',
             binary_var = True, bv_min_size = 0.1, bv_min_incindex = 0.1, other_transform = True, bounds = .98,
             transformations = ['Inv', 'Sqrt', 'Exp', 'Pw2', 'Log'], bar = False):

   
    ## check typo - #Check if the dataset contains the column name
    if x not in testdata.columns: sys.exit('ERROR!!! x variable not found in dataset, please check variable name')
    if y not in testdata.columns: sys.exit('ERROR!!! y variable not found in dataset, please check variable name')
    
    #Create duplicate columns 
    testdata['x_royalias']=testdata[x]
    testdata['y_royalias']=testdata[y]

    ## Error message
    #Exit with error for drop if missing values are more than a threshold percentage
    if sum(testdata['x_royalias'].isna())/len(testdata)>missing_thres: sys.exit('ERROR!!! x variable has more than '+str(round(missing_thres*100,1))+'% missing, please consider drop')

    #Exit with error if the column data type is not string
    if is_string_dtype(testdata['x_royalias']): sys.exit('ERROR!!! '+x+' variable is not numeric. TnI_categorical requires x variable to be numeric, please check variable type')
    
    #Exit with error for drop if there is any missing value in y
    if sum(testdata['y_royalias'].isna())>0: sys.exit('ERROR!!! y (' +y+') variable has missing value. TnI_categorical needs y variable to be fully populated. Please consider eliminating missing y values')
    
    #Exit with error if the datatype for y is a string
    if is_string_dtype(testdata['y_royalias']): sys.exit('ERROR!!! y ('+y+') variable is not numeric or integer. TnI_categorical requires y variable to be numeric or integer, please check variable type')
    
    if bar==False:
        print('method=index') if not method else print('method: '+method)## Why does bar = False; seem to be confusing    

    if method not in ['index','index_med','mean', 'median']:
        sys.exit('ERROR!!! method variable needs to be index, mean or median')


    ## missing flag
    if (missing_flag == True):
        testdata['x_missing_flag']=[1 if ct==True else 0 for ct in list(pd.isnull(testdata['x_royalias']))]

    ## core binning part
    if bar==False:
        print('min_bin_size = ',min_bin_size)

    #THIS IS FIXED
    #bnds=[1,7]
    try:
        #bnds=quantile_function_1(testdata['x_royalias'],bounds=bounds)
        bnds=quantile_function_1([x for x in list(testdata['x_royalias']) if not math.isnan(x)],bounds=bounds)
    
    except:
        print ('Inverse of Empirical Distribution Function not executed as range in beyond the interpolation range. We are instead using nearest neighbour method')
    
    finally:
        #bnds=quantile_function_2(testdata['x_royalias'],bounds=bounds)
        bnds=quantile_function_2([x for x in list(testdata['x_royalias']) if not math.isnan(x)],bounds=bounds)
    
    print(bnds)
    lookup_table_bin=pd.DataFrame({'var_name':x,'lower_bound':bnds[0],'upper_bound':bnds[1]},index=[0])
    testdata['x_royalias']=pd.Series(np.minimum(np.maximum(testdata['x_royalias'].to_numpy(),bnds[0]),bnds[1])) ## THIS NEEDS TO BE VERIFIED

    if(bnds[0] == bnds[1]):
        if bar ==False:
            print('With bounds, only one unique value')

    else:
        tab=testdata['x_royalias'].value_counts(sort=False)
        n=len(testdata['x_royalias'].isna())
        i=0
        j=0
        cut_point=[]
    
        b=tab[0:(len(tab)-1)].cumsum()/n
    
        while i<1:
            a=list(b[b>min_bin_size+i].index)
            if len(a)>=1:
                cut_point.insert(j,a[0])  ##  THIS COULD CAUSE ISSUES ON DIFFERENT DATASET
                i=b[a[0]]
                j=j+1
            else:
                i=1
                j=j-1
            
        cut_point=cut_point[0:j]

    if bar==False:
        print('bin boundaries listed below')

    cut_point=[-np.inf]+cut_point+[bnds[1]]
    if bar==False:print(cut_point)
    
    #print(testdata['x_royalias'])###################
    #print(cut_point)##################
    testdata['x_roybins_bi']=pd.cut(testdata['x_royalias'],cut_point, right=True)
    testdata['x_roybins_bi_factor']=testdata['x_roybins_bi']
    testdata['x_roybins_bi']=testdata['x_roybins_bi'].astype(str)


    #Merge mean with original dataset
    testdata=testdata.merge(testdata.groupby(['x_roybins_bi'])[['y_royalias']].mean().reset_index().rename(columns={'y_royalias':'y_royindex'}),how='left',on='x_roybins_bi')

    ## missing input and other transform part
    testdata['y_roydistance']=abs(testdata['y_royindex']-testdata['y_royindex'][testdata['x_royalias'].isna()].mean())
    if testdata['x_royalias'].isna().mean()>0:
        if method=='index':
            lookup_table_rep=pd.DataFrame({'original_name':x,'old_val':np.NaN,'new_val':testdata['x_royalias'][testdata['y_roydistance']==testdata['y_roydistance'][testdata['y_roydistance']>0].min()].mean()},index=[0])
            testdata['x_royalias'][testdata['x_royalias'].isna()]=testdata['x_royalias'][testdata['y_roydistance']==testdata['y_roydistance'][testdata['y_roydistance']>0].min()].mean()
            if bar==False:
                print('missing imputed by dependent var index based on bins. Imput value: ',testdata['x_royalias'][testdata['y_roydistance']==testdata['y_roydistance'][testdata['y_roydistance']>0].min()].mean())
    
        if method=='index_med':
            lookup_table_rep=pd.DataFrame({'original_name':x,'old_val':np.NaN,'new_val':quantile_function(testdata['x_royalias'][testdata['y_roydistance']==testdata['y_roydistance'][testdata['y_roydistance']>0].min()],.5)},index=[0])
            testdata['x_royalias'][testdata['x_royalias'].isna()]=quantile_function(testdata['x_royalias'][testdata['y_roydistance']==testdata['y_roydistance'][testdata['y_roydistance']>0].min()],.5)
            if bar==False:
                print('missing imputed by dependent var index based on bins. Imput value: ',quantile_function(testdata['x_royalias'][testdata['y_roydistance']==testdata['y_roydistance'][testdata['y_roydistance']>0].min()],.5))
    
        if method=='mean':
            lookup_table_rep=pd.DataFrame({'original_name':x,'old_val':np.NaN,'new_val':testdata['x_royalias'].mean()},index=[0])
            testdata['x_royalias'][testdata['x_royalias'].isna()]=testdata['x_royalias'].mean()
            if bar==False:
                print('missing imputed by dependent var index based on bins. Imput value: ',testdata['x_royalias'].mean())
    
        if method=='median':
            lookup_table_rep=pd.DataFrame({'original_name':x,'old_val':np.NaN,'new_val':testdata['x_royalias'].median()},index=[0])
            testdata['x_royalias'][testdata['x_royalias'].isna()]=testdata['x_royalias'].median()
            if bar==False:
                print('missing imputed by dependent var index based on bins. Imput value: ',testdata['x_royalias'].median())
    
        testdata['x_roybins_bi'][testdata['x_roybins_bi'].isna()]='missing'
        drop1=['y_roydistance','y_royalias']
    
    else:
        if bar == False:
            print('there is no missing in independent variable, no inputing was done, no missing flag was created')
        if 'med' in method:
            lookup_table_rep=pd.DataFrame({'original_name':x,'old_val':np.NaN,'new_val':testdata['x_royalias'].median()},index=[0])
        else:
            lookup_table_rep=pd.DataFrame({'original_name':x,'old_val':np.NaN,'new_val':testdata['x_royalias'].mean()},index=[0])
        drop1=['y_roydistance','y_royalias','x_royalias','x_missing_flag']

    ## take binary var part
    if binary_var==True:
        temp_x_roybins_bi=testdata['x_roybins_bi']
        testdata=pd.get_dummies(testdata,columns=['x_roybins_bi'])
        testdata['x_roybins_bi']=temp_x_roybins_bi
        del temp_x_roybins_bi
    
        drop2=[]        
        
        for ii in range((len(testdata.columns)-len(testdata['x_roybins_bi'].unique())-1),(len(testdata.columns)-1)):
            if testdata.iloc[:,ii].mean()<bv_min_size:
                drop2.append(ii)
            
            else:
                #print('incidex: '+str(abs((testdata['y_royindex'][testdata.iloc[:,ii]==1].mean()/testdata['y_royindex'].mean())-1))) ###############################
                if abs(testdata['y_royindex'][testdata.iloc[:,ii]==1].mean()/testdata['y_royindex'].mean()-1)<bv_min_incindex:
                    drop2.append(ii)
           
        if len(drop2)==0:
            if bar == False:
                print('all binary variable taken are significant')
        else:
            if bar == False:
                print(str(len(testdata['x_roybins_bi'].unique())-len(drop2))+ ' significant binary variable was taken')
                testdata=testdata.drop(list(testdata.columns[drop2]),axis='columns')
        del drop2, ii
    
    testdata=testdata[[col for col in testdata.columns if col not in ['x_roybins_bi']]]

    lookup_table=pd.DataFrame({'var_name':x, 'original_name':x, 'type':'original'},index=[0])

    ## other transform part
    if other_transform==True:
        if bar ==False:
            temp_string=""
            for i in range(0,len(transformations)):
                if i ==0:
                    temp_string=temp_string+transformations[i]
                else:
                    temp_string=temp_string+', '+transformations[i]
                
            print('continuous var transformation: '+temp_string)
            del temp_string
        
        if 'Pw2' in transformations:
            testdata['x_roytran_TnI_pw2']=testdata['x_royalias']**2
            lookup_table=lookup_table.append({'var_name':(x+'_TnI_pw2'),'original_name': x, 'type':'pw2'},ignore_index=True)
    
        if str(0) not in testdata['x_royalias'] and 'Inv' in transformations:
            testdata['x_roytran_TnI_inv']=1/testdata['x_royalias']
            lookup_table=lookup_table.append({'var_name':(x+'_TnI_inv'),'original_name': x, 'type':'inv'},ignore_index=True)
    
        if bnds[0]>0 and 'Sqrt' in transformations:
            testdata['x_roytran_TnI_sqrt']=testdata['x_royalias']**.5
            lookup_table=lookup_table.append({'var_name':(x+'_TnI_pw2'),'original_name': x, 'type':'sqrt'},ignore_index=True)


        if bnds[0]>0 and 'Log' in transformations:
            testdata['x_roytran_TnI_log']=np.log(testdata['x_royalias'])
            lookup_table=lookup_table.append({'var_name':(x+'_TnI_log'),'original_name': x, 'type':'log'},ignore_index=True)
        

        if 'Exp' in transformations:
            testdata['x_roytran_TnI_exp']=np.exp((testdata['x_royalias']-testdata['x_royalias'].min())/(testdata['x_royalias'].max()-testdata['x_royalias'].min()))
            lookup_table=lookup_table.append({'var_name':(x+'_TnI_exp'),'original_name': x, 'type':'exp'},ignore_index=True)
    
        testdata.columns=[c if 'x_roytran' not in c else c.replace('x_roytran',x) for c in testdata.columns]
    
    
    ## rename alias
    lookup_table=lookup_table.append({'var_name':(x+'_bins'),'original_name': x, 'type':'bins'},ignore_index=True)
    testdata=testdata.rename(columns={'x_roybins_bi_factor':(x+'_bins')})


    idx=[testdata.columns.get_loc(cols) for cols in testdata.columns if cols.startswith('x_roybins')]

    for ii in idx:
        rge = [testdata['x_royalias'][testdata.iloc[:,ii]==1].min(),testdata['x_royalias'][testdata.iloc[:,ii]==1].max()]
        if rge[0]==bnds[0]:
            up=-np.Inf
        else:
            up=testdata['x_royalias'][(testdata.iloc[:,ii]==0) & (testdata['x_royalias']<rge[0])].max()
    
        lookup_table_bin=lookup_table_bin.append({'var_name':testdata.columns[ii].replace('x_roybins',x),'lower_bound':up,'upper_bound':rge[1]},ignore_index=True)
        del rge, up

    lookup_table=lookup_table.append({'var_name':(x+'_TnI_processed'), 'original_name':x, 'type':'TnI_processed'},ignore_index=True)
    testdata=testdata.rename(columns={'x_royalias':(x+'_TnI_processed')})

    if testdata['x_missing_flag'].mean()>0:
        #lookup_table=lookup_table({'var_name': (x+'_missing_flag'),'original_name':x,'type':'missing_flag'},ignore_index=True)
        lookup_table=lookup_table.append({'var_name': (x+'_missing_flag'),'original_name':x,'type':'missing_flag'},ignore_index=True)        
        testdata=testdata.rename(columns={'x_missing_flag':(x+'_missing_flag')})

    for ii in idx:
        lookup_table=lookup_table.append({'var_name':testdata.columns[ii].replace('x_roybins',x), 'original_name':x,'type':'continuous binary'}, ignore_index=True)

    testdata.columns=[c.replace('x_roybins',x) if 'x_roybins' in c else c for c in testdata.columns]
    
    lookup_table=lookup_table.append({'var_name':(x+'_dependent_var_index'),'original_name':x,'type':'dependent_var_index'},ignore_index=True)

    testdata.columns=[c.replace('y_royindex',(x+'_dependent_var_index')) if 'y_royindex' in c else c for c in testdata.columns]

    testdata=testdata[[c for c in testdata.columns if c not in drop1]]
    del drop1

    testlist={'data':testdata,'lookup':{'lookup_table':lookup_table,'lookup_table_bin':lookup_table_bin,'lookup_table_rep':lookup_table_rep}}
    
    return testlist


# In[15]:


def progressbar(it, prefix="", size=60, file=sys.stdout):
    if len(it)>0:
        count = len(it)
        def show(j):
            x = int(size*j/count)
            file.write("%s[%s%s] %i/%i\r" % (prefix, "#"*x, "."*(size-x), j, count))
            file.flush()        
        show(0)
        for i, item in enumerate(it):
            yield item
            show(i+1)
        file.write("\n")
        file.flush()


# In[16]:


def TnI_smart(test_df, y, binary_min_threshold = 0.05, missing_flag = True, min_bin_size = 0.1,
              method = 'index', binary_var = True, bv_min_size = 0.1, bv_min_incindex = 0.1, missing_thres = 0.5,
              other_transform = True, bounds = .98, transformations = ['Inv', 'Sqrt', 'Exp', 'Pw2', 'Log'], bar = True):
    '''this function assumes that the y variable is numeric'''
    
    var_list=var_cat(test_df)

    #Check if dataframe is provided
    if not isinstance(test_df,pd.DataFrame): sys.exit('ERROR!!! df is not a pandas dataframe, please please convert data into a dataframe')
    dv=test_df[y]

    #pull out var lists by datatype
    int_var=var_list[var_list['data_type'] == "numeric"]
    binary_var2=var_list[var_list['data_type'] == "binary"]
    category_var=var_list[var_list['data_type'] == "categorical"] 

    #create var sets by datatype
    int_var_names=list(pd.unique(int_var['var_name']))
    binary_var_names=list(pd.unique(binary_var2['var_name']))
    category_var_names=list(pd.unique(category_var['var_name']))
    other_names=[c for c in test_df.columns if c not in (int_var_names+binary_var_names+category_var_names)]
    #Changed this so that other_names can be used from the test_df

    #for each datatype list we need to pull out there corresponding variables
    binary_df=test_df[binary_var_names]
    int_df=test_df[int_var_names]
    #int_df=int_df[,c(which(colnames(int_df)==y), which(colnames(int_df)!=y))] --- this line seems to be unnecessary
    cat_df=test_df[category_var_names]
    cat_df[y]=dv
    other_df=test_df[other_names]

    a=int_df.apply(lambda x: x.isna()).apply(lambda x: x.mean())
    b=cat_df.apply(lambda x: x.isna()).apply(lambda x: x.mean())
    #SHOULD ADD CODE FOR BINARY VARIABLE AS WELL?
    many_missing=list(a[a>missing_thres].index)+list(b[b>missing_thres].index)
    del a,b

    if len(many_missing) > 0:
        sys.exit('ERROR!!! Variable(s): '+many_missing+' has more than'+missing_thres+' missing, please consider drop')

    a_temp=cat_df.apply(lambda x: len(x.value_counts())>1000)
    many_cat=list(a_temp[a_temp==True].index)
    del a_temp

    if len(many_cat)>0:
        sys.exit('ERROR!!! '+', '.join(many_cat)+' variable is a character variable with more than 1000 levels, highly likely to be ID vars')

    if len(binary_var_names)>0:
        a=TnI_binary(binary_df,binary_min_threshold)
        binary_tni=a['data'] ###################### CHANGED TO CAPTURE BINARY FOR SCORE
        lookup_table_binary=a['lookup']['lookup_table']###################### CHANGED TO CAPTURE BINARY FOR SCORE
        lookup_table_binary_rep=a['lookup']['lookup_table_rep']###################### CHANGED TO CAPTURE BINARY FOR SCORE
        del a

    else:
        binary_tni=binary_df
        lookup_table_binary=pd.DataFrame()
        lookup_table_binary_rep=pd.DataFrame()
    
    if len(int_var_names)>1:
    #Greater than one becuase y is incuded here...
        lookup_table_cont={}
        lookup_table_cont_bin={}
        lookup_table_cont_rep={}
        if bar == False:
            for cols in [cols for cols in int_var_names if cols not in y]:
                print(cols+' variable transformation starting \n')
                tni_cont_test=TnI_cont(int_df[[y,cols]],x=cols,y=y,missing_flag=missing_flag,missing_thres=missing_thres,
                                       min_bin_size=min_bin_size,method=method,binary_var=binary_var,bv_min_size=bv_min_size,
                                       bv_min_incindex=bv_min_incindex,other_transform=other_transform,bounds=bounds,
                                       transformations=transformations,bar=bar)
                int_df=int_df.join(tni_cont_test['data'].drop([y,cols],axis=1))
                lookup_table_cont[cols]=tni_cont_test['lookup']['lookup_table']
                lookup_table_cont_bin[cols]=tni_cont_test['lookup']['lookup_table_bin']
                lookup_table_cont_rep[cols]=tni_cont_test['lookup']['lookup_table_rep']
        else:
            for cols in progressbar([cols for cols in int_var_names if cols not in y], prefix="Computing:", size=50):
                time.sleep(0.1)
                tni_cont_test=TnI_cont(int_df[[y,cols]],x=cols,y=y,missing_flag=missing_flag,missing_thres=missing_thres,
                                      min_bin_size=min_bin_size,method=method,binary_var=binary_var,bv_min_size=bv_min_size,
                                      bv_min_incindex=bv_min_incindex,other_transform=other_transform,bounds=bounds,
                                      transformations=transformations,bar=bar)
                int_df=int_df.join(tni_cont_test['data'].drop([y,cols],axis=1))
                lookup_table_cont[cols]=tni_cont_test['lookup']['lookup_table']
                lookup_table_cont_bin[cols]=tni_cont_test['lookup']['lookup_table_bin']
                lookup_table_cont_rep[cols]=tni_cont_test['lookup']['lookup_table_rep']

            lookup_table_cont=pd.concat(lookup_table_cont.values(), ignore_index=True)
            lookup_table_cont_bin=pd.concat(lookup_table_cont_bin.values(), ignore_index=True)
            lookup_table_cont_rep=pd.concat(lookup_table_cont_rep.values(), ignore_index=True)

    else:
        lookup_table_cont=pd.DataFrame()
        lookup_table_cont_bin=pd.DataFrame()
        lookup_table_cont_rep=pd.DataFrame()
    
    if len(category_var_names)>0:
        lookup_table_cat={}
        lookup_table_cat_bin={}
        lookup_table_cat_rep={}
        if bar == False:
            for cols in [cols for cols in category_var_names if cols not in y]:
                print(cols+' variable transformation starting \n')
                tni_cat_test=TnI_cat(cat_df[[y,cols]],x=cols,y=y,missing_flag=missing_flag,missing_thres=missing_thres,
                                     min_bin_size=min_bin_size,binary_var=binary_var,bv_min_size=bv_min_size,
                                     bv_min_incindex=bv_min_incindex,bar=bar)            
            
                cat_df=cat_df.join(tni_cat_test['data'].drop([y,cols],axis=1))
                lookup_table_cat[cols]=tni_cat_test['lookup']['lookup_table']
                lookup_table_cat_bin[cols]=tni_cat_test['lookup']['lookup_table_bin']
                lookup_table_cat_rep[cols]=tni_cat_test['lookup']['lookup_table_rep']
        
        else:
            for cols in progressbar([cols for cols in category_var_names if cols not in y], prefix="Computing: ", size=50):
                time.sleep(0.1)
                tni_cat_test=TnI_cat(cat_df[[y,cols]],x=cols,y=y,missing_flag=missing_flag,missing_thres=missing_thres,
                                     min_bin_size=min_bin_size,binary_var=binary_var,bv_min_size=bv_min_size,
                                     bv_min_incindex=bv_min_incindex,bar=bar)
                cat_df=cat_df.join(tni_cat_test['data'].drop([y,cols],axis=1))
                lookup_table_cat[cols]=tni_cat_test['lookup']['lookup_table']
                lookup_table_cat_bin[cols]=tni_cat_test['lookup']['lookup_table_bin']
                lookup_table_cat_rep[cols]=tni_cat_test['lookup']['lookup_table_rep']

            lookup_table_cat=pd.concat(lookup_table_cat.values(), ignore_index=True)
            lookup_table_cat_bin=pd.concat(lookup_table_cat_bin.values(), ignore_index=True)
            lookup_table_cat_rep=pd.concat(lookup_table_cat_rep.values(), ignore_index=True)

    else:
        lookup_table_cat=pd.DataFrame()
        lookup_table_cat_bin=pd.DataFrame()
        lookup_table_cat_rep=pd.DataFrame()
    
    
    cat_df=cat_df.drop([y],axis=1)
    smart_df=pd.concat([other_df, int_df, cat_df, binary_tni],ignore_index=True,axis=1)
    smart_df.columns=list(other_df.columns)+list(int_df.columns)+list(cat_df.columns)+list(binary_tni.columns)
    
    lookup_table=pd.concat([lookup_table_cont,lookup_table_cat,lookup_table_binary], ignore_index=True)############## CHANGE MADE HERE TO CAPTURE BINARY_VARIABLES
    lookup_table_rep=pd.concat([lookup_table_cont_rep,lookup_table_cat_rep,lookup_table_binary_rep], ignore_index=True) ###############CHANGE MADE HERE TO CAPTURE BINARY_VARIABLES
    lookup_table_bin=pd.concat([lookup_table_cont_bin,lookup_table_cat_bin], ignore_index=True)
    
    smart_misingflags=smart_df.filter(regex='_missing_flag')
    smart_dependent=smart_df.filter(regex='_dependent_var_index')
    smart_bins=smart_df.filter(regex='_bins')
    smart_assign=smart_df.filter(regex='_TnI_assign')

    temp_columns=smart_df.filter(regex='_TnI_assign|_missing_flag|_dependent_var_index|_bins')
    smart_predictors=smart_df[[cols for cols in smart_df.columns if cols not in temp_columns]]
    del temp_columns
    
    drop_original_columns=[cols for cols in category_var_names if cols not in y]+[cols for cols in int_var_names if cols not in y] #[cols for cols in other_names if cols not in y] This is hashed as it is to be included in the final dataframe
    smart_predictors=smart_predictors[[cols for cols in smart_predictors if cols not in drop_original_columns]]

    testlist={'smart_predictors':smart_predictors,'smart_all':smart_df,'smart_assign':smart_assign,'smart_bins': smart_bins,
              'smart_dependent':smart_dependent,'smart_misingflags':smart_misingflags,
              'lookup':{'lookup_table':lookup_table,'lookup_table_bin':lookup_table_bin,'lookup_table_rep':lookup_table_rep}}

    return testlist


# In[21]:


def is_fitted(model):
        """Checks if model object has any attributes ending with an underscore"""
        return 0 < len( [k for k,v in inspect.getmembers(model) if k.endswith('_') and not k.startswith('__')] )


# In[22]:


def TnI_score(lookup, model = None, test_data = None,test_on_model=True,feature_columns=None):
    
    '''This uses model from sklearn library; can work for any model including logistic, random forest, svm, decision trees etc
    must required: test_data, feature columns as list,test_on_model=True
    good to have: fitted model from sklearn on training dataset from TnI Smart

    '''

    if not ('lookup' in locals() or 'var' in globals()):
        sys.exit('TnI lookup object is needed.')
    if all([False if x in list(lookup.keys()) else True for x in ['lookup_table', 'lookup_table_bin', 'lookup_table_rep']]):
        sys.exit('lookup needs to be the lookup table result from TnI.')

    lookup_table=lookup['lookup_table']
    lookup_table_bin=lookup['lookup_table_bin']
    lookup_table_rep=lookup['lookup_table_rep']
    
    #The code below tries to find the variables to be transformed
    count=0
    try:
        if test_on_model==False:
            count=1
        elif test_on_model==True:   
            if is_fitted(model)==False:
                sys.exit('Model is not fitted; please fit the model and come back again to use TnI score')
    
            else:
                try:
                    temp_type=lookup_table['type'].map(lambda x: x not in ['assign_bins', 'bins', 'dependent_var_index', 'original'])
                    temp_original=lookup_table['original_name'].map(lambda x: x in list(test_data.columns))
                    temp_var_name=lookup_table['var_name'][temp_type & temp_original]
                    model_var=[cols for cols in feature_columns if cols in list(temp_var_name)]                
                    del temp_type,temp_original,temp_var_name
        
                except:
                    warnings.warn('Test dataset is empty so checking difference in columns between features_columns and lookup table')
                    try:
                        temp_type=lookup_table['type'].map(lambda x: x not in ['assign_bins', 'bins', 'dependent_var_index', 'original'])
                        temp_var_name=lookup_table['var_name'][temp_type]
                        model_var=[cols for cols in feature_columns if cols in list(temp_var_name)]                    
                        del temp_type,temp_original,temp_var_name
                    except:
                        sys.exit('Error in format: check the format of lookup table for TnI output or feature_columns as a list of columns')
        
                model_no_var=[cols for cols in feature_columns if cols not in model_var]
                if len(model_no_var)>0:
                    warnings.warn('Variables: '+str(",".join(model_no_var))+' are not in the lookup table or original variable name are not in the data set.')
                   
    except:
        count=1
        pass


    if count==1:
        try:
            if test_data.empty==False:
                temp_type=lookup_table['type'].map(lambda x: x not in ['assign_bins', 'bins', 'dependent_var_index', 'original'])
                temp_original=lookup_table['original_name'].map(lambda x: x in list(test_data.columns))
                temp_var_name=lookup_table['var_name'][temp_type & temp_original]
                model_var=temp_var_name
                del temp_type,temp_original,temp_var_name
                count=0
        
        except:
            count=2
            pass

    if count==2:
        try:
            temp_type=lookup_table['type'].map(lambda x: x not in ['assign_bins', 'bins', 'dependent_var_index', 'original'])
            model_var=lookup_table['var_name'][temp_type]
            del temp_type
    
        except:
            sys.exit('Could no identify variables that require transformation')

    
    #The code below does the main transformation
    n=len(model_var)
    count=0
    del_colmns=[]
    try:
        if test_data.empty==False:
            if n>0:
                for i in range(0,n):
                    var_name=list(model_var)[i]            
                    original_name=list(lookup_table['original_name'][lookup_table['var_name']==var_name].values)[0]
                    del_colmns.append(original_name)
                    type_var=list(lookup_table['type'][lookup_table['var_name']==var_name].values)[0]
            
                    if not type_var=='original':
                        if type_var=='binary binary':
                            ### THIS CODE CANNOT HANDLE THRESHOLD AS DONE IS BINARY_TNI
                            old_val=list(lookup_table_rep['old_val'][lookup_table_rep['original_name']==original_name].values)
                            dat=test_data[original_name]
                            dat=list(pd.Series(dat).fillna('missing'))
                            dat=[str(x) for x in dat]
                            idx=[old_val.index(x)+1 if x in old_val else None for x in dat]
                            dat=idx
    
                        else: 
                            old_val=list(lookup_table_rep['old_val'][lookup_table_rep['original_name']==original_name].values)
                            new_val=lookup_table_rep['new_val'][lookup_table_rep['original_name']==original_name]
                            dat=test_data[original_name]
                            if len(old_val) > 0:
                                idx=[old_val.index(x)+1 if x in old_val else None for x in dat]
                                idx2=[i for i, n in enumerate(idx) if n is not None]
                                for i in idx2:
                                    dat.iloc[i]=new_val[idx[i]]
                    
                        pd.Series(dat).astype('float')    
                        if type_var=='sqrt':
                            test_data[var_name]=np.sqrt(np.abs(dat))
                
                        elif type_var=='pw2':
                            test_data[var_name]=dat**2 
                
                        elif type_var=='TnI_processed':
                            test_data[var_name]=dat
                
                        elif type_var in 'assign':
                            test_data[var_name]=dat
                    
                        elif type_var=='exp':
                            test_data[var_name]=np.exp((dat-lower_bound)/(upper_bound-lower_bound))

                        elif type_var=='inv':
                            #THIS REQUIRED CONDITION THAT dat>0
                            test_data[var_name]=1/dat
                
                        elif type_var=='log':
                            #IS THIS CORRECT? ABS can change the whole value
                            test_data[var_name]=np.log(np.abs(dat))
                
                        elif type_var=='missing_flag':
                            test_data[var_name]=test_data[original_name].isna().astype('int')
                
                        elif type_var in 'categorical binary':
                            upper_bound=lookup_table_bin['upper_bound'][lookup_table_bin['var_name']==var_name].iloc[0]
                            lower_bound=lookup_table_bin['lower_bound'][lookup_table_bin['var_name']==var_name].iloc[0]
                            test_data[var_name]=list(((dat<=upper_bound) & (dat>lower_bound)).map(lambda x: int(x)).values)
                    
                        elif type_var=='continuous binary':
                            upper_bound=lookup_table_bin['upper_bound'][lookup_table_bin['var_name']==original_name].iloc[0]
                            lower_bound=lookup_table_bin['lower_bound'][lookup_table_bin['var_name']==original_name].iloc[0]
                            dat=np.minimum(np.maximum(dat,lower_bound),upper_bound)
                            upper_bound=lookup_table_bin['upper_bound'][lookup_table_bin['var_name']==var_name].iloc[0]
                            lower_bound=lookup_table_bin['lower_bound'][lookup_table_bin['var_name']==var_name].iloc[0]
                            test_data[var_name]=list(((dat<=upper_bound) & (dat>lower_bound)).map(lambda x: int(x)).values)
                    
                        elif type_var=='binary binary':
                            dummy_col=var_name.split('_')[len(var_name.split('_'))-1]
                            new_dummy_col_val=list(lookup_table_rep['new_val'][(lookup_table_rep['old_val']==dummy_col) & (lookup_table_rep['original_name']==original_name)])[0]
                            test_data[var_name]=[1 if x==new_dummy_col_val else 0 for x in dat]
                    
            test_data=test_data[[cols for cols in test_data.columns if cols not in del_colmns]]
            test_data=test_data[feature_columns]
    
            if test_on_model==True:
                try:
                    test_data['TnI_score']=model.predict(test_data)
                    return test_data 
        
                except:
                    sys.exit('You want to predict but prediction is not executed!! Check you model!')
                    count=33
                    pass
            else:
                test_data=test_data
                return test_data
        
    except:
        sys.exit('Transaformation detected error! Check you dataframes or model for issues!')

